//
// Smile Decorder V8 4808N Next18 Rev.A exp.04
// Copyright(C)'2016- Nagoden / Desktop Station / Fujigaya2 / MECY 
//
// exp.03 2023/07/06 :Change the logic of CV3 and CV4 to the same logic as SmileSoundDecoder 
//
//


#include <Arduino.h>
#include <string.h>
#include <avr/pgmspace.h> 
//#include <avr/eeprom.h>
#include "motor_func.h"
#include "motor_ctrl.h"
//#include <SoftwareSerial.h>
#include "NmraDcc_x.h"
/*************************************************
 * 
 * Declaration of Symbols
 * 
 *************************************************/

//#define DEBUG			//リリースのときはコメントアウトすること
#define MAN_VER_NUMBER  04   /* Release Ver CV07 */
#define MAN_ID_NUMBER 108  /* Manufacture ID */


#define DECODER_ADDRESS 3

//Pin Func  Port
//24  AUX1  A1
//25  F0_F  A2
//26  F0_R  A3
//27  AUX2  A4
//28  AUX5  A5
//30  AUX6  D0
// 9  AUX4  D5
//10  AUX3  D6
#define PIN_AUX1	A1
#define PIN_F0_F	A2
#define PIN_F0_R	A3
#define PIN_AUX2	A4
#define PIN_AUX5	A5
#define PIN_AUX6  A6
#define PIN_AUX4  5
#define PIN_AUX3  6

// Define the Arduino Pin to connect to the DCC input signal
#define DCC_PIN 2

// ATMEGA4808 / ARDUINO
//
//                 (4)  (3)  (2)                (RST) (A7)
//                 PA2  PA1  PA0  GND  VDD UPDI  PF6  PF5
//
//                 32   31   30   29   28   27   26   25
//              + ____ ____ ____ ____ ____ ____ ____ ____  +
//   (5) PA3   1|                                         |24  PF4 (A6)
//   (6) PA4   2|                                         |23  PF3 (A5)
//   (7) PA5   3|                                         |22  PF2 (A4)
//   (8) PA6   4|                                         |21  PF1 (RX)
//   (9) PA7   5|                32pin QFN                |20  PF0 (TX)
//  (11) PC0   6|                                         |19  GND
//  (12) PC1   7|                                         |18  AVDD
//  (13) PC2   8|                                         |17  PD7 (25)
//               + ____ ____ ____ ____ ____ ____ ____ ____ +
//                   9   10   11   12   13   14   15   16 
//
//                  PC3  PD0  PD1  PD2  PD3  PD4  PD5  PD6
//                 (10) (A0) (A1) (A2) (A3) (22) (23) (24)

//Use jtag2updi for sketch writing Connect PD6 and UPDI


#define MOTOR_LIM_MAX 255

#define CV_VSTART 2              //スタート電圧
#define CV_ACCRATIO 3            //加速率
#define CV_DECCRATIO  4          //減速率
#define CV_VMAX 5                //最高電圧
#define CV_DECODERMODE  30       //デコーダモード
#define CV_AUTO_DIMMING 50       //ヘッドライト自動減光
#define CV_HEAD_DIMMING  51      //ヘッドライト減光調整
#define CV_HEAD_INTENSITY 52     //ヘッドライト光量調整
#define CV_TAIL_INTENSITY 53     //テールライト光量調整
#define CV_BEMFcoefficient  54   //BEMF効果度合
#define CV_PI_P 55               //BEMF Pパラメータ
#define CV_PI_I 56               //BEMF Iパラメータ
#define CV_BEMFCUTOFF 57         //BEMF カットオフ値
#define CV_AUX1_INTENSITY 58     //AUX1（室内灯）光量調整

#define CV_FUNC_F0 110           //F00 Function
#define CV_FUNC_F1 111           //F01 Function
#define CV_FUNC_F2 112           //F02 Function
#define CV_FUNC_F3 113           //F03 Function
#define CV_FUNC_F4 114           //F04 Function
#define CV_FUNC_F5 115           //F05 Function
#define CV_FUNC_F6 116           //F06 Function
#define CV_FUNC_F7 117           //F07 Function
#define CV_FUNC_F8 118           //F08 Function
#define CV_FUNC_F9 119           //F09 Function
#define CV_FUNC_F10 120           //F10 Function
#define CV_FUNC_F11 121           //F11 Function
#define CV_FUNC_F12 122           //F12 Function
#define CV_FUNC_F13 123           //F13 Function
#define CV_FUNC_F14 124           //F14 Function
#define CV_FUNC_F15 125           //F15 Function
#define CV_FUNC_F16 126           //F16 Function
#define CV_FUNC_F17 127           //F17 Function
#define CV_FUNC_F18 128           //F18 Function
#define CV_FUNC_F19 129           //F19 Function
#define CV_FUNC_F20 130           //F20 Function
#define CV_FUNC_F21 131           //F21 Function
#define CV_FUNC_F22 132           //F22 Function
#define CV_FUNC_F23 133           //F23 Function
#define CV_FUNC_F24 134           //F24 Function
#define CV_FUNC_F25 135           //F25 Function
#define CV_FUNC_F26 136           //F26 Function
#define CV_FUNC_F27 137           //F27 Function
#define CV_FUNC_F28 138           //F28 Function
#define CV_FUNCSTATE 200

/*************************************************
 * 
 * Declaration of Classes
 * 
 *************************************************/

NmraDcc	 Dcc;
DCC_MSG	 Packet;

/*************************************************
 * 
 * create servo object to control a servo 
 * 
 *************************************************/

/*************************************************
 * 
 * Declaration of Variables
 * 
 *************************************************/


//Task Schedule
unsigned long gPreviousL1 = 0;
unsigned long gPreviousL2 = 0;
unsigned long gPreviousL3 = 0;
unsigned long gPreviousL4 = 0;
unsigned long gPreviousL5 = 0; //161117 fujigaya2
unsigned long gPreviousL6 = 0; //170419 fujigaya2
unsigned long gIntervalRail = 1000;
unsigned long gElapsed_time = 0;

//モータ制御関連の変数

uint16_t gSpeedCmd = 0;
uint16_t gSpeedCmd_vvvf = 0;
uint16_t gSpeedCmd_rail = 0;//fujigaya2
uint8_t gPwmDirv = 128;
uint8_t gCV29_Vstart = 0;
uint8_t gCV29Direction = 0;
uint8_t gDecoderMode = 0; //2018/01/01　nagoden


// CV

uint8_t gCV1_SAddr = 3;
uint8_t gCVx_LAddr = 3;
uint8_t gCV2_Vstart = 16;
uint8_t gCV3_AccRatio = 0;
uint8_t gCV4_DecRatio = 0;
uint8_t gCV5_VMAX = 250;
uint8_t gCV30_DecoderMode = 0;  //デコーダモード　0：モータ　1:モータ制御なし　2018/01/01　nagoden

uint8_t gCV50_AUTO_DIMMING = 0;      //ヘッドライト自動減光
uint8_t gCV51_HEAD_DIMMING = 255;    //ヘッドライト減光調整
uint8_t gCV52_HEAD_INTENSITY = 255;  //ヘッドライト光量調整
uint8_t gCV53_TAIL_INTENSITY = 255;  //テールライト光量調整

uint8_t gCV54_BEMFcoefficient = 64;   // 0-127 step / Reajust by MECY 2017/10/08
uint8_t gCV55_PI_P = 18;              // Reajust by MECY 2017/10/08
uint8_t gCV56_PI_I = 96;
uint8_t gCV57_BEMF_CutOff = 0;        // 0でBEMF Off 1でBEMF On
uint8_t gCV58_AUX1_INTENSITY = 255;  //AUX1（室内灯）光量調整

//ファンクションの変数
uint8_t gF0_FCONTROL = 0;     //ヘッド・テールのライトライト制御
uint8_t gF0_BCONTROL = 0;     //ヘッド・テールのライトライト制御
uint8_t gHEAD_DIMMING = 0;    //ヘッドライト減光スイッチ
uint8_t gF0_SW = 0;    //消灯モードスイッチ
uint8_t gFuncBits[29] = {0};
uint8_t gCVFuncTable[29] = {0};
uint8_t gAUTO_DIMMING_FLG = 0;  //自動減光フラグ

struct CVPair{
  uint16_t	CV;
  uint8_t	Value;
};

CVPair FactoryDefaultCVs [] = {
  {CV_MULTIFUNCTION_PRIMARY_ADDRESS, DECODER_ADDRESS},
  {CV_ACCESSORY_DECODER_ADDRESS_MSB, 0},		//The LSB is set CV 1 in the libraries .h file, which is the regular address location, so by setting the MSB to 0 we tell the library to use the same address as the primary address. 0 DECODER_ADDRESS
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB, 0},	 //XX in the XXYY address
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB, 0},	 //YY in the XXYY address
  {CV_29_CONFIG, 2},	 //Make sure this is 0 or else it will be random based on what is in the eeprom which could caue headaches
//  {CV_29_CONFIG, 128},   //Make sure this is 0 or else it will be random based on what is in the eeprom which could caue headaches
  {CV_VSTART, 16},
  {CV_ACCRATIO, 0},
  {CV_DECCRATIO, 0},
  {CV_VMAX, 250},
  {CV_DECODERMODE, 0},
  {CV_AUTO_DIMMING, 0},
  {CV_HEAD_DIMMING, 255},
  {CV_HEAD_INTENSITY, 255},
  {CV_TAIL_INTENSITY, 255},  
  {CV_BEMFcoefficient, 64},
  {CV_PI_P, 18},
  {CV_PI_I, 96},
  {CV_BEMFCUTOFF,0},
  {CV_AUX1_INTENSITY, 255},
  {CV_FUNC_F0, 1},  //Light Control  
  {CV_FUNC_F1, 2},  //Room Light Control
  {CV_FUNC_F2 + 0, 0},
  {CV_FUNC_F2 + 1, 0},
  {CV_FUNC_F2 + 2, 0},
  {CV_FUNC_F2 + 3, 0},
  {CV_FUNC_F2 + 4, 0},
  {CV_FUNC_F2 + 5, 0},
  {CV_FUNC_F2 + 6, 0},
  {CV_FUNC_F2 + 7, 0},
  {CV_FUNC_F2 + 8, 0},
  {CV_FUNC_F2 + 9, 0},
  {CV_FUNC_F2 + 10, 0},
  {CV_FUNC_F2 + 11, 0},
  {CV_FUNC_F2 + 12, 0},
  {CV_FUNC_F2 + 13, 0},
  {CV_FUNC_F2 + 14, 0},
  {CV_FUNC_F2 + 15, 0},
  {CV_FUNC_F2 + 16, 0},
  {CV_FUNC_F2 + 17, 0},
  {CV_FUNC_F2 + 18, 0},
  {CV_FUNC_F2 + 19, 0},
  {CV_FUNC_F2 + 20, 0},
  {CV_FUNC_F2 + 21, 0},
  {CV_FUNC_F2 + 22, 0},
  {CV_FUNC_F2 + 23, 0},
  {CV_FUNC_F2 + 24, 0},
  {CV_FUNC_F2 + 25, 0},
  {CV_FUNC_F2 + 26, 0},
	{200 + 0, 0}, // Function status
	{200 + 1, 0},
	{200 + 2, 0},
	{200 + 3, 0},
	{200 + 4, 0},
	{200 + 5, 0},
	{200 + 6, 0},
	{200 + 7, 0},
	{200 + 8, 0},
	{200 + 9, 0},
	{200 + 10, 0},
	{200 + 11, 0},
	{200 + 12, 0},
	{200 + 13, 0},
	{200 + 14, 0},
	{200 + 15, 0},
	{200 + 16, 0},
	{200 + 17, 0},
	{200 + 18, 0},
	{200 + 19, 0},
	{200 + 20, 0},
	{200 + 21, 0},
	{200 + 22, 0},
	{200 + 23, 0},
	{200 + 24, 0},
	{200 + 25, 0},
	{200 + 26, 0},
	{200 + 27, 0},
	{200 + 28, 0}
};

/*************************************************
 * 
 * Declaration of Functions
 * 
 *************************************************/

void(* resetFunc) (void) = 0;  //declare reset function at address 0

uint8_t FactoryDefaultCVIndex = sizeof(FactoryDefaultCVs) / sizeof(CVPair);

void ControlRoomLight(void);
void ControlHeadLight(void);


/*************************************************
 * 
 * Functions
 * 
 *************************************************/

void notifyCVResetFactoryDefault()
{
	//When anything is writen to CV8 reset to defaults. 

  resetCVToDefault();	
   
  #ifdef DEBUG
    Serial.println("Resetting...");
  #endif
  delay(1000);  //typical CV programming sends the same command multiple times - specially since we dont ACK. so ignore them by delaying
  resetFunc();
}


void resetCVToDefault()
{
	//Reset CV to Default
  #ifdef DEBUG
	  Serial.println("CVs being reset to factory defaults");
  #endif
	MOTOR_Ack();
	for (int j=0; j < FactoryDefaultCVIndex; j++ ){
		Dcc.setCV( FactoryDefaultCVs[j].CV, FactoryDefaultCVs[j].Value);
	}
}

extern void	   notifyCVChange( uint16_t CV, uint8_t Value)
{
   //Message when CV is changed
  #ifdef DEBUG
    Serial.print("CV "); 
    Serial.print(CV); 
    Serial.print(" Changed to "); 
    Serial.println(Value, DEC);
  #endif
}

void notifyCVAck(void)
{
  #ifdef DEBUG
    Serial.println("notifyCVAck");
  #endif
 
	MOTOR_Ack();
}

void setup()
{
	// ソフトウエアシリアル通信レートセット:
	  Serial.begin(9600);

	//ファンクションの割り当てピン初期化

	  pinMode(PIN_F0_F, OUTPUT);
  	digitalWrite(PIN_F0_F, 0);
  	pinMode(PIN_F0_R, OUTPUT);
    digitalWrite(PIN_F0_R, 0);	
    pinMode(PIN_AUX1, OUTPUT);
    digitalWrite(PIN_AUX1, 0);
	  pinMode(PIN_AUX2, OUTPUT);
	  digitalWrite(PIN_AUX2, 0);
	  pinMode(PIN_AUX5, OUTPUT);
	  digitalWrite(PIN_AUX5, 0);
    pinMode(PIN_AUX6, OUTPUT);
    digitalWrite(PIN_AUX6, 0);

	  if ( Dcc.getCV(CV_MULTIFUNCTION_PRIMARY_ADDRESS) == 0xFF )
	  {
		   //if eeprom has 0xFF then assume it needs to be programmed
      #ifdef DEBUG
	      Serial.println("CV Defaulting due to blank eeprom");
      #endif
	    notifyCVResetFactoryDefault();
    } 
   else{
      #ifdef DEBUG
	      Serial.println("CV Not Defaulting");
      #endif
    }
	// Setup which External Interrupt, the Pin it's associated with that we're using, disable pullup.
	    Dcc.pin(DCC_PIN,0);

	// Call the main DCC Init function to enable the DCC Receiver
	    Dcc.init( MAN_ID_NUMBER, MAN_VER_NUMBER,   FLAGS_MY_ADDRESS_ONLY , 0 ); 

	//Reset task
	  gPreviousL1 = millis();
	  gPreviousL2 = millis();
	  gPreviousL3 = millis();
	  gPreviousL4 = millis();
	  gPreviousL5 = millis();
	//Init CVs
  	gCV1_SAddr = Dcc.getCV( CV_MULTIFUNCTION_PRIMARY_ADDRESS ) ;
  	gCV2_Vstart = Dcc.getCV( CV_VSTART ) ;
  	gCV3_AccRatio = Dcc.getCV( CV_ACCRATIO ) ;
  	gCV4_DecRatio = Dcc.getCV( CV_DECCRATIO ) ;
	  gCV5_VMAX = Dcc.getCV( CV_VMAX ) ;
	  gCVx_LAddr = (Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB ) << 8) + Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB );
  	gCV29_Vstart = Dcc.getCV( CV_29_CONFIG ) ;
    gCV30_DecoderMode = Dcc.getCV( CV_DECODERMODE ) ; // add 2018/01/01 nagoden
    gDecoderMode = gCV30_DecoderMode;                 // add 2023/07/10 nagoden
    gCV50_AUTO_DIMMING = Dcc.getCV( CV_AUTO_DIMMING ) ;  //ヘッドライト自動減光
    gCV51_HEAD_DIMMING = Dcc.getCV( CV_HEAD_DIMMING ) ;    //ヘッドライト減光調整
    gCV52_HEAD_INTENSITY = Dcc.getCV( CV_HEAD_INTENSITY ) ;  //ヘッドライト光量調整
    gCV53_TAIL_INTENSITY = Dcc.getCV( CV_TAIL_INTENSITY ) ;  //テールライト光量調整

	  gCV54_BEMFcoefficient =  Dcc.getCV( CV_BEMFcoefficient ) ;
	  gCV55_PI_P = Dcc.getCV( CV_PI_P ) ;
	  gCV56_PI_I = Dcc.getCV( CV_PI_I ) ;
	  gCV57_BEMF_CutOff = Dcc.getCV( CV_BEMFCUTOFF ) ;
    gCV58_AUX1_INTENSITY = Dcc.getCV( CV_AUX1_INTENSITY ) ;  //AUX1（室内灯）光量調整
 //AUX3,4, Pin assignment
    pinMode(PIN_AUX3, OUTPUT);
    digitalWrite(PIN_AUX3, 0);
    pinMode(PIN_AUX4, OUTPUT);
    digitalWrite(PIN_AUX4, 0);

	/* ファンクションテーブル初期化・読み取り */

	  for( int i = 0; i < 29; i++)
	  {
		  gCVFuncTable[i ] = Dcc.getCV( CV_FUNC_F0 + i ) ;

	    #ifdef DEBUG
      Serial.print("CV ");
      Serial.print(CV_FUNC_F0 + i);
      Serial.print(" = ");
      Serial.println( gCVFuncTable[i ] );
      #endif
  
		
		/* 異常値は0に固定 */
  		if( gCVFuncTable[i] > 10)
	  	{
		  	gCVFuncTable[i] = 0;
  	  
		  }
	  }
  
	  //ファンクション状態をロード
	  for( int i = 0; i < 29; i++)
	  {
		  gFuncBits[i] = Dcc.getCV( CV_FUNCSTATE + i ) ;
			/* 異常値は0に固定 */
		  if( gFuncBits[i] > 1)
	  	{
		  	gFuncBits[i] = 0;
		  }
	  }

	//cv29 Direction Check
	  if ( (gCV29_Vstart & 0x01) > 0)
	  {
		  gCV29Direction = 1; //REVをFWDにする
	  }
	  else
	  {
		  gCV29Direction = 0;//FWDをFWDにする
	  }

  
// Initialization of motor control related
	  MOTOR_Init();
	  MOTOR_SetCV(2, gCV2_Vstart);
	  MOTOR_SetCV(3, gCV3_AccRatio);
	  MOTOR_SetCV(4, gCV4_DecRatio);
	  MOTOR_SetCV(5, gCV5_VMAX);
	  MOTOR_SetCV(54, gCV54_BEMFcoefficient);   // Nagoden Adjust1 2016/06/02
	  MOTOR_SetCV(55, gCV55_PI_P);				      // MECY Adjust1 2016/04/09
	  MOTOR_SetCV(56, gCV56_PI_I);				      // MECY Adjust1 2016/04/09
	  MOTOR_SetCV(57, gCV57_BEMF_CutOff);


    #ifdef DEBUG
	    Serial.print("CV1(ShortAddr): ");
    	Serial.println(gCV1_SAddr);
	    Serial.print("CV17/18(LongAddr): ");
	    Serial.println(gCVx_LAddr);
	    Serial.print("CV2(Vstart): ");
	    Serial.println(gCV2_Vstart);
    	Serial.print("CV3(AccRatio): ");
    	Serial.println(gCV3_AccRatio);
    	Serial.print("CV4(DecRatio): ");
    	Serial.println(gCV4_DecRatio);
	    Serial.print("CV5(VoltageMAX): ");
    	Serial.println(gCV5_VMAX);
      Serial.print("CV29(Config): ");
      Serial.println(gCV29_Vstart);
      Serial.print("CV30(DecoderMode): ");
      Serial.println(gCV30_DecoderMode);
      Serial.print("CV50 (AUTO_DIMMING): ");
      Serial.println(gCV50_AUTO_DIMMING);
      Serial.print("CV51 (HEAD_DIMMING): ");
      Serial.println(gCV51_HEAD_DIMMING);
      Serial.print("CV52 (HEAD_INTENSITY): ");
      Serial.println(gCV52_HEAD_INTENSITY);
      Serial.print("CV53 (TAIL_INTENSITY): ");
      Serial.println(gCV53_TAIL_INTENSITY);
    	Serial.print("CV54 (BEMF coefficient): ");
	    Serial.println(gCV54_BEMFcoefficient);
	    Serial.print("CV55 (PI P Gain): ");
	    Serial.println(gCV55_PI_P);
	    Serial.print("CV56 (PI I Gain): ");
	    Serial.println(gCV56_PI_I);
	    Serial.print("CV57 (BEMF Cut Off): ");
	    Serial.println(gCV57_BEMF_CutOff);
    	Serial.println("Ready");
    #endif

}

void loop()
{
	
	// You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation
	Dcc.process();
	
	if( (millis() - gPreviousL1) >= 10)
	{
  		//Speed detection
	  	MOTOR_Sensor();
		//Reset task
		gPreviousL1 = millis();
	}	

	if( (millis() - gPreviousL2) >= 500)  //ヘッドライト自動減光待ち時間
	{
//    if ((gSpeedCmd <= 0) && (gCV50_AUTO_DIMMING != 0)) {
    if ((gPWMRef <= 0) && (gCV50_AUTO_DIMMING != 0)) {      //加速減速率の変更によりPWMの速度命令gPWMRefを参照に変更　Rev03 Nagoden
      
      gElapsed_time = gElapsed_time + 500 ;
      if (gElapsed_time >= (gCV50_AUTO_DIMMING*1000)) {
        gAUTO_DIMMING_FLG = 1;  //減光
      }
    }
    else {
      gAUTO_DIMMING_FLG = 0;
      gElapsed_time = 0;  
    }
    if (gElapsed_time >= 256000){   //オーバフロー回避
      gElapsed_time = 0;  //タイムカウントリセット
    }
		gPreviousL2 = millis();

	} 
	
	if( (millis() - gPreviousL3) >= 100)
	{
 //   if (gCV30_DecoderMode != 1 ) //2018/01/01 Nagoden  Rev04 Delete
 //   {
		  //Motor drive control
		  MOTOR_Main(gSpeedCmd, gPwmDirv);
 //   }
		//Reset task
		gPreviousL3 = millis();   
	}
	
	if( (millis() - gPreviousL4) >= 250)
	{	
		//Reset task
		gPreviousL4 = millis();
	}

  if( (millis() - gPreviousL5) >= gIntervalRail)
  { 
    //Reset task
    gPreviousL5 = millis();
  }

  if( (millis() - gPreviousL6) >= 100)
  {
    //Reset task
    gPreviousL6 = millis();
  }
  
}



//DCC速度信号の受信によるイベント
extern void notifyDccSpeed( uint16_t Addr, DCC_ADDR_TYPE AddrType, uint8_t Speed, DCC_DIRECTION Dir, DCC_SPEED_STEPS SpeedSteps )
{
	uint16_t aSpeedRef = 0;
	//速度値の正規化(255を100%とする処理)
	if( Speed >= 1)
	{
		aSpeedRef = ((Speed - 1) * 255) / SpeedSteps;
	}
	else
	{
		//緊急停止信号受信時の処理 //Nagoden comment 2016/06/11
    #ifdef DEBUG
		    Serial.println("***** Emagency STOP **** ");
    #endif
		aSpeedRef = 0;
  }
  gSpeedCmd = aSpeedRef;

  gPwmDirv = Dir;
  if ( gCV29Direction > 0)
  {
     if ( Dir == DCC_DIR_FWD) {
          gPwmDirv = 128;
     }
    else  {
          gPwmDirv = 0;     
    }     
  }
  else  {
     if ( Dir == DCC_DIR_FWD) {
          gPwmDirv = 0;
     }
    else {
         gPwmDirv =128;     
    }     
  }

  #ifdef DEBUG
	  // デバッグメッセージ
	  Serial.print("Speed - ADR: ");
	  Serial.print(Addr);
	  Serial.print(", SPD: ");
	  Serial.print(Speed);
	  Serial.print(", DIR: ");
	  Serial.print(Dir);
	  Serial.print(", aSpeedRef: ");
	  Serial.println(aSpeedRef);
  #endif

}

//ファンクション信号受信のイベント
extern void notifyDccFunc( uint16_t Addr, DCC_ADDR_TYPE AddrType, FN_GROUP FuncGrp, uint8_t FuncState)
{
	switch(FuncGrp)
	{
    case FN_0_4:
//      Function F00
         if  ((FuncState & FN_BIT_00) == 16)
         {
            if (gFuncBits[0] == 0)
            {
//               Dcc.setCV(CV_FUNCSTATE + 0, 1);
            }
             gFuncBits[0] = 1;
             exec_function (gFuncBits[0], gCVFuncTable[0] );
         }    
         else
         {
            if (gFuncBits[0] == 1) 
            {
              gFuncBits[0] = 0;
              exec_function ( gFuncBits[0], gCVFuncTable[0] );
  //    		    Dcc.setCV(CV_FUNCSTATE + 0, 0);
            }
         }
             
//      Function F01
         if  ((FuncState & FN_BIT_01) == 1){
            if (gFuncBits[1] == 0)
            {              
              gFuncBits[1] = 1;
              exec_function ( gFuncBits[1], gCVFuncTable[1] );
//              Dcc.setCV(CV_FUNCSTATE + 1, 1);
            }
         }    
         else
         {
            if (gFuncBits[1] == 1) 
            {
              gFuncBits[1] = 0;
              exec_function ( gFuncBits[1], gCVFuncTable[1] );
//              Dcc.setCV(CV_FUNCSTATE + 1, 0);
            }
         }
//      Function F02
         if  ((FuncState & FN_BIT_02) == 2)
	    	 {
           gFuncBits[2] = 1;
           exec_function (gFuncBits[2] , gCVFuncTable[2] );
         }    
         else 
          {
           gFuncBits[2] = 0;
           exec_function ( gFuncBits[2], gCVFuncTable[2] ); 
         }
//      Function F03
         if  ((FuncState & FN_BIT_03) == 4)  
         {
           gFuncBits[3] = 1;
           exec_function ( gFuncBits[3], gCVFuncTable[3] );
          }
		 	    else
		     	{
           gFuncBits[3] = 0;
           exec_function ( gFuncBits[3], gCVFuncTable[3] );
		 	    }
         
//      Function F04 
         if ((FuncState & FN_BIT_04) == 8)
         {
           gFuncBits[4] = 1;
           exec_function ( gFuncBits[4], gCVFuncTable[4] );
         }
         else
         {
           gFuncBits[4] = 0;
           exec_function ( gFuncBits[4], gCVFuncTable[4] );
         } 
      break;
    case FN_5_8:
//      Function F05    
         if  ((FuncState & FN_BIT_05) == 1)
         {
           gFuncBits[5] = 1;
           exec_function ( gFuncBits[5], gCVFuncTable[5] );
         }       
         else
         {
           gFuncBits[5] = 0;
           exec_function ( gFuncBits[5], gCVFuncTable[5] );
         } 
//      Function F06    
         if  ((FuncState & FN_BIT_06) == 2) 
         {
            gFuncBits[6] = 1;
            exec_function ( gFuncBits[6], gCVFuncTable[6] );
         }    
         else 
         {
            gFuncBits[6] = 0;
            exec_function ( gFuncBits[6], gCVFuncTable[6] );
         }
//      Function F07    
         if  ((FuncState & FN_BIT_07) == 4) 
         {
            gFuncBits[7] = 1;
            exec_function ( gFuncBits[7], gCVFuncTable[7] );
         }    
         else 
         {
            gFuncBits[7] = 0;
            exec_function ( gFuncBits[7], gCVFuncTable[7] );
         }
//      Function F08    
         if  ((FuncState & FN_BIT_08) == 8)
         {
            gFuncBits[8] = 1;
            exec_function ( gFuncBits[8], gCVFuncTable[8] );
         }    
         else 
         {
            gFuncBits[8] = 0;
            exec_function ( gFuncBits[8], gCVFuncTable[8] );
	        }
      break;
    case FN_9_12:
//      Function F09    
         if  ((FuncState & FN_BIT_09) == 1) 
         {
            gFuncBits[9] = 1;
            exec_function ( gFuncBits[9], gCVFuncTable[9] );
         }    
         else 
         {
            gFuncBits[9] = 0;
            exec_function ( gFuncBits[9], gCVFuncTable[9] );
         }
//      Function F10    
         if  ((FuncState & FN_BIT_10) == 2) 
         {
            gFuncBits[10] = 1;
            exec_function ( gFuncBits[10], gCVFuncTable[10] );
         }    
         else 
         {
            gFuncBits[10] = 0;
            exec_function ( gFuncBits[10], gCVFuncTable[10] );
         }
//      Function F11    
         if  ((FuncState & FN_BIT_11) == 4) 
         {
            gFuncBits[11] = 1;
            exec_function ( gFuncBits[11], gCVFuncTable[11] );
         }    
         else 
         {
            gFuncBits[11] = 0;
            exec_function ( gFuncBits[11], gCVFuncTable[11] );
         }

//      Function F12    
         if  ((FuncState & FN_BIT_12) == 8) 
         {
            gFuncBits[12] = 1;
            exec_function ( gFuncBits[12], gCVFuncTable[12] );
         }    
         else 
         {
            gFuncBits[12] = 0;
            exec_function ( gFuncBits[12], gCVFuncTable[12] );
         }

      break;

    case FN_13_20:
//      Function F13
         if  ((FuncState & FN_BIT_13) == 1) 
         {
            gFuncBits[13] = 1;
            exec_function ( gFuncBits[13], gCVFuncTable[13] );
         }    
         else 
         {
            gFuncBits[13] = 0;
            exec_function ( gFuncBits[13], gCVFuncTable[13] );
         }
//      Function F14    
         if  ((FuncState & FN_BIT_14) == 2) 
         {
            gFuncBits[14] = 1;
            exec_function ( gFuncBits[14], gCVFuncTable[14] );
         }    
         else 
         {
            gFuncBits[14] = 0;
            exec_function ( gFuncBits[14], gCVFuncTable[14] );
         }
//      Function F15    
         if  ((FuncState & FN_BIT_15) == 4) 
         {
            gFuncBits[15] = 1;
            exec_function ( gFuncBits[15], gCVFuncTable[15] );
         }    
         else 
         {
            gFuncBits[15] = 0;
            exec_function ( gFuncBits[15], gCVFuncTable[15] );
         }
//      Function F16    
         if  ((FuncState & FN_BIT_16) == 8) 
         {
            gFuncBits[16] = 1;
            exec_function ( gFuncBits[16], gCVFuncTable[16] );
         }    
         else 
         {
            gFuncBits[16] = 0;
            exec_function ( gFuncBits[16], gCVFuncTable[16] );
         }
//      Function F17    
         if  ((FuncState & FN_BIT_17) == 16) 
         {
            gFuncBits[17] = 1;
            exec_function ( gFuncBits[17], gCVFuncTable[17] );
         }    
         else 
         {
            gFuncBits[17] = 0;
            exec_function ( gFuncBits[17], gCVFuncTable[17] );
         }
//      Function F18    
         if  ((FuncState & FN_BIT_18) == 32) 
         {
            gFuncBits[18] = 1;
            exec_function ( gFuncBits[18], gCVFuncTable[18] );
         }    
         else 
         {
            gFuncBits[18] = 0;
            exec_function ( gFuncBits[18], gCVFuncTable[18] );
         }
//      Function F19    
         if  ((FuncState & FN_BIT_19) == 64) 
         {
            gFuncBits[19] = 1;
            exec_function ( gFuncBits[19], gCVFuncTable[19] );
         }    
         else 
         {
            gFuncBits[19] = 0;
            exec_function ( gFuncBits[19], gCVFuncTable[19] );
         }

 //      Function F20    
        if  ((FuncState & FN_BIT_20) == 128) 
        {
            gFuncBits[20] = 1;
            exec_function ( gFuncBits[20], gCVFuncTable[20] );
         }    
         else 
         {
            gFuncBits[20] = 0;
            exec_function ( gFuncBits[20], gCVFuncTable[20] );
         }
      break;
      
    case FN_21_28:
//      Function F21    

         if  ((FuncState & FN_BIT_21) == 1) 
         {
            gFuncBits[21] = 1;
            exec_function ( gFuncBits[21], gCVFuncTable[21] );
         }    
         else 
         {
            gFuncBits[21] = 0;
            exec_function ( gFuncBits[21], gCVFuncTable[21] );
         }
//      Function F22    
         if  ((FuncState & FN_BIT_22) == 2) 
         {
            gFuncBits[22] = 1;
            exec_function ( gFuncBits[22], gCVFuncTable[22] );
         }    
         else 
         {
            gFuncBits[22] = 0;
            exec_function ( gFuncBits[22], gCVFuncTable[22] );
         }
//      Function F23    
         if  ((FuncState & FN_BIT_23) == 4) 
         {
            gFuncBits[23] = 1;
            exec_function ( gFuncBits[23], gCVFuncTable[23] );
         }    
         else 
         {
            gFuncBits[23] = 0;
            exec_function ( gFuncBits[23], gCVFuncTable[23] );
         }
//      Function F24    
         if  ((FuncState & FN_BIT_24) == 8) 
         {
            gFuncBits[24] = 1;
            exec_function ( gFuncBits[24], gCVFuncTable[24] );
         }    
         else 
         {
            gFuncBits[24] = 0;
            exec_function ( gFuncBits[24], gCVFuncTable[24] );

         }
//      Function F25    
         if  ((FuncState & FN_BIT_25) == 16) 
         {
            gFuncBits[25] = 1;
            exec_function ( gFuncBits[25], gCVFuncTable[25] );
         }    
         else 
         {
            gFuncBits[25] = 0;
            exec_function ( gFuncBits[25], gCVFuncTable[25] );
         }
//      Function F26    
         if  ((FuncState & FN_BIT_26) == 32) 
         {
            gFuncBits[26] = 1;
            exec_function ( gFuncBits[26], gCVFuncTable[26] );
         }    
         else 
         {
            gFuncBits[26] = 0;
            exec_function ( gFuncBits[26], gCVFuncTable[26] );
         }
//      Function F27    
         if  ((FuncState & FN_BIT_27) == 64) 
         {
            gFuncBits[27] = 1;
            exec_function ( gFuncBits[27], gCVFuncTable[27] );
         }    
         else 
         {
            gFuncBits[27] = 0;
            exec_function ( gFuncBits[27], gCVFuncTable[27] );
         }
 //      Function F28    
         if  ((FuncState & FN_BIT_28) == 128) 
         {
            gFuncBits[28] = 1;
            exec_function ( gFuncBits[28], gCVFuncTable[28] );
         }    
         else 
         {
            gFuncBits[28] = 0;
            exec_function ( gFuncBits[28], gCVFuncTable[28] );
         }
     break;
	}
  
}

//exec_function   
// FuncState　OFF=0 ON=1
// FuncCV     0:OFF 1:F0  2:AUX1  3:AUX2  4:AUX3  5:AUX4  6:AUX5  7:AUX6  8:ヘッドライト減光 9:F0消灯 
//   FuncCVはCV111から138までの値
void exec_function ( int FuncState, int FuncCV )
{
 switch( FuncCV )
        {
        case 0:  // F0の処理
          break ;

         case 1: // F1の処理
            ControlHeadLight(FuncState);
            break ;
         case 2: // AUX1の処理
          if(FuncState == 0) {
            analogWrite(PIN_AUX1, 0);
          }
          else  {
            analogWrite(PIN_AUX1, gCV58_AUX1_INTENSITY);
          }
          break ;
        case 3: // AUX2の処理
          if(FuncState == 0) {
            digitalWrite(PIN_AUX2, 0);
          }
          else  {
            digitalWrite(PIN_AUX2, 1);
          }
          break ;
        case 4: // AUX3の処理
          if(FuncState == 0) {
            digitalWrite(PIN_AUX3, 0);
          }
          else  {
            digitalWrite(PIN_AUX3, 1);
          }
          break ;
        case 5: // AUX4の処理
          if(FuncState == 0) {
            digitalWrite(PIN_AUX4, 0);
          }
          else  {
            digitalWrite(PIN_AUX4, 1);
          }
          break ;
        case 6: // AUX5の処理
          if(FuncState == 0) {
            digitalWrite(PIN_AUX5, 0);
          }
          else  {
            digitalWrite(PIN_AUX5, 1);
          }
          break ;
        case 7: // AUX6の処理
          if(FuncState == 0) {
            digitalWrite(PIN_AUX6, 0);
          }
          else  {
            digitalWrite(PIN_AUX6, 1);
          }
          break ;
        case 8: // ヘッドライト減光
          if(FuncState == 0) {
            gHEAD_DIMMING = 0; 
          }
          else  {
            gHEAD_DIMMING = 1; 
          }
          break ;
        case 9: // F0消灯モード
          if(FuncState == 0) {
            gF0_SW = 0;
          }
          else  {
            gF0_SW = 1;
          }
          break ;
    }
}

void ControlHeadLight(int FuncState)
{
//	if(gFuncBits[0] == 0)
  if(FuncState == 0)   //2020/11/22  Nagoden
	{
      analogWrite(PIN_F0_F, 0);
      analogWrite(PIN_F0_R, 0);
		return;
	}
	if( gPwmDirv == 0)
	{
		//FWD
    if (gF0_SW == 0) {
      if (gAUTO_DIMMING_FLG == 1){
        analogWrite(PIN_F0_F, gCV51_HEAD_DIMMING);
      }
      else{
        if (gHEAD_DIMMING == 1) { 
          analogWrite(PIN_F0_F, gCV51_HEAD_DIMMING);
        }
        else{
          analogWrite(PIN_F0_F, gCV52_HEAD_INTENSITY);
        }
      }
    }  
    else{
      analogWrite(PIN_F0_F, 0);
    }     
    analogWrite(PIN_F0_R, 0);
    return;
	}
	else
	{
		//REV
    if (gF0_SW == 0) {
      analogWrite(PIN_F0_R, gCV53_TAIL_INTENSITY);
    }
    else{
      analogWrite(PIN_F0_R, 0);
    }  
      analogWrite(PIN_F0_F, 0);

      return;
   
	}
}
