/*
// Motor Control for DS-DCC decode
// By yaasan
// Rev02(nagoden) 
//    Added inversion processing for PWM output duty value by changing motor driver to MP6513.
//
*/
#include <Arduino.h>
#include <string.h>
#include "motor_ctrl.h"
#include "motor_func.h"


/*************************************************
 * 
 * Declaration of Symbols
 * 
 *************************************************/

#define COUNT_ZEROMAX 10


/*************************************************
 * 
 * Declaration of Classes
 * 
 *************************************************/


/*************************************************
 * 
 * Declaration of Variables
 * 
 *************************************************/


//モータ制御関連の変数
long gMotorLPF_buf = 0;
long gBEMFLPF_buf = 0;

long gMotorPI_buf = 0;
long gMotorSpeed = 0;
byte gDirection = 128;
int gPrevSpeed = 0;
int gPrevSpeedRef = 0;
int gPWMRef = 0;
byte gZeroCount = 0;

MOTOR_PARAM gParam;

byte gAckDir = 0;

//HLR用
bool gHLR_running = false;
int32_t gHLR_SpeedRef = 0;
int32_t gHLR_LastSpeed = 0;
unsigned long gHLR_Time = 0;
int32_t gHLR_AccTime = 2000;//加減速時間(単位は1ms/d)
int32_t gHLR_DecTime = 1000;//加減速時間(単位は1ms/d)
int32_t gHLR_DecTimeEmg = 1000;//加減速時間(単位は1ms/d)
int32_t gHLR_TargetTime;//HLR時に使用する変化速度幅に合わせた遷移時間

//モード
bool gHLR_EmgMode = 0;

/*************************************************
 * 
 * Declaration of Functions
 * 
 *************************************************/

void MOTOR_Init();
void MOTOR_SetCV(byte iNo, byte inData);
void MOTOR_Sensor();
void MOTOR_Ack(void);
void MOTOR_Main(int inSpeedCmd, int inDirection);

void MOTOR_clearHLR();
int32_t MOTOR_calcHLR(int32_t inSpeedRef);


/*************************************************
 * 
 * Functions
 * 
 *************************************************/

void MOTOR_Init()
{
	gParam.mStartVoltage = 32;
	gParam.mBEMFcoefficient = 38;
	gParam.mPI_P = 32;
	gParam.mPI_I = 96;
 	gParam.mBEMFcutoff = 32;
	gParam.mAccRatio = 4 * 16;
	gParam.mDecRatio = 4 * 16;
	
//ATMega4809(Bit0はEnable-Disable）
//Value  Name    数値    周波数  倍率
//0x0   DIV1    0b0001  62.9KHz 1
//0x1   DIV2    0b0011  31.4KHz 2.0
//0x2   DIV4    0b0101  15.7KHz 4.0
//0x3   DIV8    0b0111  7.86KHz 8.0
//0x4   DIV16   0b1001  3.93KHz 16.0
//0x5   DIV64   0b1011  983Hz   64.0
//0x6   DIV256  0b1101  246Hz   256
//0x7   DIV1024 0b1111  61.4Hz  1024
TCA0.SINGLE.CTRLA = 0b0011 ;

	//PWM出力ピン D22,D23を出力にセット
	pinMode(MOTOR_PWM_B, OUTPUT);
	pinMode(MOTOR_PWM_A, OUTPUT);
}


void MOTOR_SetCV(byte iNo, byte inData)
{
	switch(iNo)
	{
	case 2:
		gParam.mStartVoltage = inData;
		break;

	case 3: // 加速率CV3
		
		if( inData > 255)
		{
			gHLR_AccTime = 255 * 600;//20ms*30d=0.6sec, 0.6*255=153sec = 2min 33sec.
		}
		else
		{
			gHLR_AccTime =  (int32_t)inData * 600;//20ms*30d=0.6sec, 
		}
		
#ifdef DEBUG
		Serial.print("AccTime: ");
		Serial.println(gHLR_AccTime);
		
#endif
		
		break;

	case 4: // 減速率CV4
		if( inData > 255)
		{
			gHLR_DecTime = 255 * 600;//20ms*30d=0.6sec, 0.6*255=153sec = 2min 33sec.
			gHLR_DecTimeEmg = gHLR_DecTime / 2;
		}
		else
		{
			gHLR_DecTime =  (int32_t)inData * 600;//20ms*30d=0.6sec, 
			gHLR_DecTimeEmg = gHLR_DecTime / 2;
		}
		
#ifdef DEBUG
		Serial.print("DecTime: ");
		Serial.println(gHLR_DecTime);
		
#endif
		break;
		
	case 5:
		gParam.mMaxVoltage = inData;
		break;
	case 54:
    	gParam.mBEMFcoefficient = inData;
    	break;
	case 55:
		gParam.mPI_P = inData;
		break;
	case 56:
		gParam.mPI_I = inData;
		break;
 	case 57:
  	gParam.mBEMFcutoff = inData;
    	
/*** 利用停止 MECY 2017/8/6
    	if(0 < inData < 20)
		{
			gParam.mBEMFcutoff = 20;                  // Adjust CV57 Para set by MECY 2017/4/16
		}
		else
		{
			gParam.mBEMFcutoff = inData;              // Adjust CV57 Para set by MECY 2017/4/16
		}
***/		
    	break;

	}
	
	
}

void MOTOR_Ack(void)
{
//  MP6513で方向のトグルを実行すると破損の可能性があるので片側のみの対応とする  2020/03/01 Nagoden
//	if( gAckDir == 0)
//	{
		analogWrite(MOTOR_PWM_B, 255);    //Change  by MP6513.
		analogWrite(MOTOR_PWM_A, 30);     //Change  by MP6513.
//		gAckDir = 1;
//	}
//	else
//	{
//		analogWrite(MOTOR_PWM_B, 30);     //Change  by MP6513.
//		analogWrite(MOTOR_PWM_A, 255);    //Change  by MP6513.
//		gAckDir = 0;
//	}
	delay( 6 );  
	analogWrite(MOTOR_PWM_A, 255);      //Change  by MP6513.
	analogWrite(MOTOR_PWM_B, 255);      //Change  by MP6513.
}


void MOTOR_Sensor()
{
	//int aBEMF_result, aSpeed_calculated;
	int aBEMF_result = 0 ;           // BEMF enhance by MECY 2017/09/17
	int aSpeed_calculated = 0;       // BEMF enhance by MECY 2017/09/17
	
	//BEMF detection
	aBEMF_result = MOTOR_GetBEMF();
	
	aSpeed_calculated = aBEMF_result /6 ;    // BEMF enhance by MECY 2017/09/24
	
	if(aSpeed_calculated > 255)
	{
		aSpeed_calculated = 255;
	}
	
	//Speed conversion
	gMotorSpeed = MOTOR_LPF(aSpeed_calculated, 8, &gBEMFLPF_buf);
	
	
}


/* Motor control Task (10Hz) */
void MOTOR_Main(int inSpeedCmd, int inDirection)
{
	
	uint32_t aSpeedRef = (uint32_t)inSpeedCmd * (uint32_t)(gParam.mMaxVoltage  - gParam.mStartVoltage) / (uint32_t)gParam.mMaxVoltage ;
	int aPWMRef = 0;
	long aSpeedRef_offseted = 0;

	//SmileSound互換HLR
	if( (gHLR_AccTime > 0) && (gHLR_DecTime > 0) )
	{
		//加減速時間考慮した加減速制御
		aSpeedRef = MOTOR_calcHLR(aSpeedRef);
	}
	else
	{
		//8bit LPF
		if( gPrevSpeedRef <= aSpeedRef)
		{
			aSpeedRef = MOTOR_LPF(aSpeedRef, gParam.mAccRatio, &gMotorLPF_buf); // 加速
		}
		else
		{
			aSpeedRef = MOTOR_LPF(aSpeedRef, gParam.mDecRatio, &gMotorLPF_buf); // 減速
		}
	}
	
	//バックアップ
	//gPrevSpeedRef = inSpeedCmd;
    gPrevSpeedRef = aSpeedRef;
	
	//PI Control
	if( gParam.mBEMFcutoff > 0)
	{
		
		//Speed detection
		int aSpeedDet = (gMotorSpeed * (127 - (int)gParam.mBEMFcoefficient)) / 127;     // 127 step value  /Adjust MECY 2016/12/24
		
		if( aSpeedDet < gParam.mBEMFcutoff)
		{
			//ある閾値以下は速度ゼロと見なす
			aSpeedDet = 0;
		}
		else
		{
			//検出速度のオフセットをする
			aSpeedDet = aSpeedDet - gParam.mBEMFcutoff;
		}
		
		//Command zero check
		if( aSpeedDet == 0)
		{
			gZeroCount++;
			
			if( gZeroCount > COUNT_ZEROMAX)
			{
				gZeroCount = 0;
				gMotorPI_buf = 0;
			}
		}
		else
		{
			gZeroCount = 0;	
		}
		
			
		//BEMF時 PI制御
		if((aSpeedRef == 0) && ( aSpeedDet == 0))
		{
			aSpeedRef_offseted = MOTOR_PI(aSpeedRef - aSpeedDet, gParam.mPI_P, gParam.mPI_I, &gMotorPI_buf) ;
		}
		else if((aSpeedRef >= 1) && ( aSpeedDet >= 0))
		{
			aSpeedRef_offseted = MOTOR_PI(aSpeedRef - aSpeedDet, gParam.mPI_P, gParam.mPI_I, &gMotorPI_buf) + gParam.mStartVoltage ; //CV2 response  /MECY ReAdjust 2017/08/27
		}
	}
	else
	{
		//PI無効
		if( aSpeedRef == 0)
		{
			aSpeedRef_offseted = aSpeedRef;
		}
		else if( aSpeedRef >= 1)
		{
			aSpeedRef_offseted = aSpeedRef + gParam.mStartVoltage ; //2016/06/19 Modify Nagoden 
		}
	}

	//前回速度として保存
	gPrevSpeed = aSpeedRef_offseted ;

	//Limiter (8bit PWM Output)
	aPWMRef = MOTOR_limitSpeed((uint16_t)(aSpeedRef_offseted));

  //Clear output and LPF buffer when direction changes  2019/08/10 Nagoden
  if ( inDirection != gDirection )
  {
       aPWMRef = 0;
       gMotorLPF_buf = 0;
  }     
  
	//PWM出力

  if ( gDecoderMode != 1 )
  {
	  if( aPWMRef == 0)
	  {
		  analogWrite(MOTOR_PWM_A, 255);    //Change  by MP6513.
		  analogWrite(MOTOR_PWM_B, 255);    //Change  by MP6513.
    //Turn off the output for a certain time when the direction is changed 2019/08/10 Nagoden
      if ( inDirection != gDirection )
      {
          delay(1000); 
          gDirection = inDirection;
      }     
	  }
	  else
	  {
      //Max Value check  Change  by MP6513.
      if( aPWMRef >= 255)
      {
       aPWMRef = 255;
      }
 		  //PWM出力
		  //進行方向でPWMのABを切り替える
		  if( inDirection > 0)
		  {
  		    analogWrite(MOTOR_PWM_B, 255);              //Change  by MP6513.
	  	    analogWrite(MOTOR_PWM_A, 255 - aPWMRef);    //Change  by MP6513.
	    }
	    else
	    {
	  	  analogWrite(MOTOR_PWM_A, 255);             //Change  by MP6513.
	  	  analogWrite(MOTOR_PWM_B, 255 - aPWMRef);   //Change  by MP6513.
	    }
	  }
  }

	gDirection = inDirection;
	gPWMRef = aPWMRef;
	
	/*
	Serial.print("SCMD= ");
	Serial.print(inSpeedCmd);

	Serial.print(" ,SREF= ");
	Serial.print(aSpeedRef);
	
	Serial.print(" ,BEMF= ");
	Serial.print(aSpeedDet);

	Serial.print(" ,PI_IBUF= ");
	Serial.print(gMotorPI_buf);

	Serial.print(" ,PWM=");
	Serial.println(aPWMRef);
	*/
	
}

int MOTOR_GetBEMF()
{
  if( gParam.mBEMFcutoff > 0)
  {
	uint32_t aAve = 0;
	int i;
	
	//Stop PWM
	analogWrite(MOTOR_PWM_B, 255);  //Change  by MP6513.
	analogWrite(MOTOR_PWM_A, 255);  //Change  by MP6513.
	
	delayMicroseconds(300);
	
	for( i = 0; i < 8; i++)
	{
		aAve = aAve + analogRead(A0);
		delayMicroseconds(30);
	}

  //MaxValueCheck
    if( gPWMRef >= 255)
    {
        gPWMRef = 255;
    }
  
	//進行方向でPWMのABを切り替える
	if( gDirection > 0){
		analogWrite(MOTOR_PWM_B, 255);            //Change  by MP6513.
		analogWrite(MOTOR_PWM_A, 255 - gPWMRef);  //Change  by MP6513.
	}
	else
	{
		analogWrite(MOTOR_PWM_A, 255);            //Change  by MP6513.
		analogWrite(MOTOR_PWM_B, 255 - gPWMRef);  //Change  by MP6513.
	}
	
	return (aAve >> 3);
	
  }
}



int MOTOR_limitSpeed(int inSpeed)
{
	uint16_t aSpeedref = 0;                                   //MECY Adjust 2016/06/04
	
	if( inSpeed >= gParam.mMaxVoltage)
	{
		aSpeedref = gParam.mMaxVoltage;                            //MECY Adjust 2016/06/04
	}
	else if( inSpeed <= 0)
	{
		aSpeedref = 0;                                        //MECY Adjust 2016/06/04
	}
	else if( gParam.mBEMFcutoff == 0 )                        //MECY Adjust 2016/06/04
	{
		aSpeedref = inSpeed;
	}
	else if( gParam.mBEMFcutoff > 0 )
	{
		aSpeedref = inSpeed;
	}
	
	return aSpeedref;
}



void MOTOR_clearHLR()
{
	gHLR_running = false;
	gHLR_SpeedRef = 0;
	gHLR_TargetTime = 0;
}

int32_t MOTOR_calcHLR(int32_t inSpeedRef)
{
	int32_t aSpeedRef = inSpeedRef;
	int32_t aAccDecTime;
	float aSpeedCalcTemp = 0;
	float aTimeCalcTemp = 0;
	float aDiffTime;
	
	if( gHLR_running == true)
	{
		//加減速制御中
		
		//終了判定
		if( (millis() - gHLR_Time) >= gHLR_TargetTime)
		{
			//ホフラフ到達
			gHLR_running = false;
			
			//速度演算の更新の必要無し(入力値inSpeedRefをaSpeedRefに設定しており、到達速度であるため。)
		}
		else
		{
			//加減速中のとき
			
			//速度指令値をなだらかに変化させる。
			aDiffTime = millis() - gHLR_Time;
			
			//現在速度の計算(演算誤差に影響があるので一部float化)
			aSpeedCalcTemp = ((float)((gHLR_SpeedRef - gHLR_LastSpeed) * (aDiffTime)) / ((float)gHLR_TargetTime)) + (float)gHLR_LastSpeed;
			
			//キャストし、固定小数点に戻す
			aSpeedRef = (int32_t)aSpeedCalcTemp;
			
			if( inSpeedRef != gHLR_SpeedRef)
			{
				//現在速度を前回速度に変更し、
				gHLR_LastSpeed = aSpeedRef;
				gHLR_SpeedRef = inSpeedRef;
				
				gHLR_Time = millis();
				aAccDecTime = ( gHLR_LastSpeed < gHLR_SpeedRef) ? gHLR_AccTime : (gHLR_EmgMode == true ? gHLR_DecTimeEmg : gHLR_DecTime);
				
				//演算誤差をなくすためにfloat化
				aTimeCalcTemp = ((float)aAccDecTime * (float)ABS(gHLR_SpeedRef - gHLR_LastSpeed)) / 255.0;
				
				//キャストし、intに戻す
				gHLR_TargetTime = (int32_t)aTimeCalcTemp;
				
				if( gHLR_TargetTime == 0)
				{
					//ERROR
					gHLR_TargetTime = 1;
				}
				
			}
		
		}
	}
	else
	{
		
		//速度指令値の変更を取得
		if( inSpeedRef != gHLR_SpeedRef)
		{
			//現在速度を前回速度に変更し、
			gHLR_LastSpeed = gHLR_SpeedRef;
			gHLR_SpeedRef = inSpeedRef;
			gHLR_Time = millis();
			aAccDecTime = ( gHLR_LastSpeed < gHLR_SpeedRef) ? gHLR_AccTime : gHLR_DecTime;
			
			
			//演算誤差をなくすためにfloat化
			aTimeCalcTemp = ((float)aAccDecTime * (float)ABS(gHLR_SpeedRef - gHLR_LastSpeed)) / 255.0;
			
			//キャストし、intに戻す
			gHLR_TargetTime = (int32_t)aTimeCalcTemp;
			
			//加速させない
			aSpeedRef = gHLR_LastSpeed;
			
			if( gHLR_TargetTime == 0)
			{
				//ERROR
				gHLR_TargetTime = 1;
			}
			
			
			
			gHLR_running = true;
		}
		
	}
	
	//HLR後速度指令値
	return aSpeedRef;
}
